#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mpi.h"

#define MPIR_ALLTOALL_SHORT_MSG 256
#define MPIR_ALLTOALL_MEDIUM_MSG 32768

#define MPIR_ALLTOALL_TAG 9
#define MPIR_ALLTOALLV_TAG 10

#if defined (_XPMPI)
int XPMPI_Alltoall(void * sendbuf, int sendcount, MPI_Datatype sendtype,
     	  	   void * recvbuf, int recvcount, MPI_Datatype recvtype,
		   MPI_Comm comm)
#else
int MPI_Alltoall(void * sendbuf, int sendcount, MPI_Datatype sendtype,
     		 void * recvbuf, int recvcount, MPI_Datatype recvtype,
		 MPI_Comm comm)
#endif
{
	int comm_size, rank, i;
	int src, dest, nbytes, pof2;
	MPI_Aint send_extent, recv_extent;
	MPI_Request * reqarray;
	MPI_Status * starray, status;

	PMPI_Comm_size(comm, &comm_size);
	PMPI_Comm_rank(comm, &rank);
	/* Fix this up */

	PMPI_Type_extent(sendtype, &send_extent);
	PMPI_Type_extent(recvtype, &recv_extent);
	nbytes = send_extent * sendcount;

	reqarray = (void *) malloc(2 * comm_size * sizeof(MPI_Request));
	starray = (void *) malloc(2 * comm_size * sizeof(MPI_Status));

	if (nbytes <= MPIR_ALLTOALL_SHORT_MSG) {
		printf("Very small sizes not supported yet\n");
		PMPI_Abort(MPI_COMM_WORLD, -1);
	}
	else if (nbytes <= MPIR_ALLTOALL_MEDIUM_MSG) {
		printf("Medium message algorithm\n");
		for (i = 0; i < comm_size; i++) {
			dest = (rank + i) % comm_size;
#if defined (_MPI)
			MPI_Irecv((char *) recvbuf + dest * recvcount * recv_extent,
#else
			PMPI_Irecv((char *) recvbuf + dest * recvcount * recv_extent,
#endif
				   recvcount, recvtype, dest, MPIR_ALLTOALL_TAG, comm,
				   &reqarray[i]);
		}

		for (i = 0; i < comm_size; i++) {
			dest = (rank + i) % comm_size;
#if defined (_MPI)
			MPI_Isend((char *) sendbuf + dest * sendcount * send_extent,
#else
			PMPI_Isend((char *) sendbuf + dest * sendcount * send_extent,
#endif
				  sendcount, sendtype, dest, MPIR_ALLTOALL_TAG, comm,
				  &reqarray[i + comm_size]);
		}

#if defined (_MPI)
		MPI_Waitall(2 * comm_size, reqarray, starray);
#else
		PMPI_Waitall(2 * comm_size, reqarray, starray);
#endif
		for (i = 0; i < 2 * comm_size; i++) {
			if (starray[i].MPI_ERROR != MPI_SUCCESS) {
				printf("Wait error\n");
				PMPI_Abort(MPI_COMM_WORLD, -1);
			}
		}
	}
	else {
		printf("Large message algorithm\n");
		memcpy(recvbuf + rank * sendcount * send_extent,
		       sendbuf + rank * recvcount * recv_extent,
		       (sendcount * send_extent));

		i = 1;
		while (i < comm_size) i *= 2;
		if (i == comm_size) pof2 = 1; else pof2 = 0;

		for (i = 1; i < comm_size; i++) {
			if (pof2) {
				src = dest = rank ^ i;
			}
			else {
				src = (rank - i + comm_size) % comm_size;
				dest = (rank + i) % comm_size;
			}

#if defined (_MPI)
			MPI_Sendrecv((char *) sendbuf + dest * sendcount * send_extent,
#else
			PMPI_Sendrecv((char *) sendbuf + dest * sendcount * send_extent,
#endif
				     sendcount, sendtype, dest, MPIR_ALLTOALL_TAG,
				     (char *) recvbuf + src * recvcount * recv_extent,
				     recvcount, recvtype, src, MPIR_ALLTOALL_TAG, comm, &status);
		}
	}

	return MPI_SUCCESS;
}


#if defined(_XPMPI)
int XPMPI_Alltoallv(void * sendbuf, int * sendcnts, int * sdispls, MPI_Datatype sendtype,
		    void * recvbuf, int * recvcnts, int * rdispls, MPI_Datatype recvtype,
		    MPI_Comm comm)
#else
int MPI_Alltoallv(void * sendbuf, int * sendcnts, int * sdispls, MPI_Datatype sendtype,
		  void * recvbuf, int * recvcnts, int * rdispls, MPI_Datatype recvtype,
		  MPI_Comm comm)
#endif
{
	int src, dest, rank, comm_size;
	int i, req_count;
	MPI_Aint send_extent, recv_extent;
	MPI_Request * reqarray;
	MPI_Status * starray, status;

	PMPI_Comm_size(comm, &comm_size);
	PMPI_Comm_rank(comm, &rank);

	PMPI_Type_extent(sendtype, &send_extent);
	PMPI_Type_extent(recvtype, &recv_extent);

	reqarray = (void *) malloc(2 * comm_size * sizeof(MPI_Request));
	starray = (void *) malloc(2 * comm_size * sizeof(MPI_Status));

	req_count = 0;
	for (i = 0; i < comm_size; i++) {
		dest = (rank + i) % comm_size;
#if defined (_MPI)
		if (MPI_Irecv((char *) recvbuf + rdispls[dest] * recv_extent,
#else
		if (PMPI_Irecv((char *) recvbuf + rdispls[dest] * recv_extent,
#endif
			       recvcnts[dest], recvtype, dest, MPIR_ALLTOALLV_TAG, comm,
			       &reqarray[req_count]) != MPI_SUCCESS) {
			printf("Irecv error\n");
			PMPI_Abort(MPI_COMM_WORLD, -1);
		}
		req_count++;
	}

	for (i = 0; i < comm_size; i++) {
		dest = (rank + i) % comm_size;
#if defined (_MPI)
		if (MPI_Isend((char *) sendbuf + sdispls[dest] * send_extent,
#else
		if (PMPI_Isend((char *) sendbuf + sdispls[dest] * send_extent,
#endif
			      sendcnts[dest], sendtype, dest, MPIR_ALLTOALLV_TAG, comm,
			      &reqarray[req_count]) != MPI_SUCCESS) {
			printf("Isend error\n");
			PMPI_Abort(MPI_COMM_WORLD, -1);
		}
		req_count++;
	}

#if defined (_MPI)
	if (MPI_Waitall(req_count, reqarray, starray) == MPI_ERR_IN_STATUS) {
#else
	if (PMPI_Waitall(req_count, reqarray, starray) == MPI_ERR_IN_STATUS) {
#endif
		for (i = 0; i < req_count; i++) {
			if (starray[i].MPI_ERROR != MPI_SUCCESS) {
				printf("Wait error\n");
				PMPI_Abort(MPI_COMM_WORLD, -1);
			}
		}
	}

	return MPI_SUCCESS;
}
