/*#####################################################################
#                                                                   #
#      (c) COPYRIGHT IBM CORP. 2006 ALL RIGHTS RESERVED.             #
#                                                                   #
#                LICENSED MATERIALS-PROPERTY OF IBM.                #
#                                                                   #
#####################################################################
*/

/* 
$Id   : bgl3dfft.hpp
*/

/** *  \addtogroup FFTS
 *  The BGL3DFFT class computes the complex to complex 3D FFT. The input data is 
 *  distributed on a 3D processor mesh. Each processor has (Ni/Pi)^3 elements, where i=0,1,2 (x,y,z)
 *  It is templated on the sign (FORWARD or REVERSE) 
 *  and the data type (float or double).
 *  The 1D FFT performs the 1D FFTs locally on each node.
 **/

#ifndef __BGL_3DFFT_H_
#define __BGL_3DFFT_H_
// Follows the essl sign convention
#define FORWARD 1  
#define REVERSE -1

template <class T> class PrivateFFT;
template <class T> class complexType;

template< int FWD_REV, class T>
class BGL3DFFT
{
private:
  PrivateFFT<T> *privateFFT;  
  void PrintVector(complexType<T>* data, int size, int myGlobalID,  int dir);

public:  
  enum{ DIM = 4 };
  // Initialize the BGL3DFFT to perform 3D FFT on data of size globalNx X globalNy X globalNz
  // on pX x pY x Pz processor mesh
  void Init(const int globalNx, const int globalNy, const int globalNz, const int pX, const int pY, const int pZ);
  void Finalize();
  
  // aLocalSizeX, aLocalSizeY, aLocalSiseZ, are data sizes along each axis on each processor
  // aLocalRecipSizeX, aLocalRecipSizeY, aLocalRecipSizeZ are the data sizes along each axis on each processor after 3D FFT.
  // aLocalX, aLocalY, aLocalZ, are the data offsets before the 3 DFFT.
  // aLocalRecipX, aLocalRecipY, aLocalRecipZ, are the data offsets after the 3 DFFT.
  
  void GetLocalDimensions( int& aLocalX, int& aLocalSizeX,
			   int& aLocalY, int& aLocalSizeY,
			   int& aLocalZ, int& aLocalSizeZ,
			   int& aLocalRecipX, int& aLocalRecipSizeX,
			   int& aLocalRecipY, int& aLocalRecipSizeY,
			   int& aLocalRecipZ, int& aLocalSizeRecipZ );
  
  // This function performs the 3D FFT for forward and reverse.
  // For the Forward FFT:
  // bglfftIn is local input data of size aLocalSizeX \times aLocalSizeY \times aLocalSizeZ
  //          Data stride: z is the faster and x is the slower
  // bglfftOut is local output data of size aLocalRecipSizeX \times aLocalRecipSizeY \times aLocalRecipSizeZ
  //          Data stride: x is the faster and z is the slower
  
  // For the Reverse FFT:
  // bglfftIn is the local input data of size aLocalRecipSizeX \times aLocalRecipSizeY \times aLocalRecipSizeZ
  //          Data stride: x is the faster and z is the slower
  
  // bglfftOut is the local out data of size aLocalSizeX \times aLocalSizeY \times aLocalSizeZ
  //          Data stride: z is the faster and x is the slower  
  
  void DoFFT(complexType<T>* bglfftIn, complexType<T>* bglfftOut) ; 
};

#endif
