
! This sample program illustrates the 
! use of P3DFFT library for highly scalable parallel 3D FFT. 
!
! This program initializes a 3D array with a 3D sine wave, then 
! performs forward transform, backward transform, and checks that 
! the results are correct, namely the same as in the start except 
! for a normalization factor. It can be used both as a correctness
! test and for timing the library functions. 
!
! Processor grid in this test is set up as to minimize P1.
!
! To build this program, use one of the provided makefiles, modifying 
! it as needed for your compilers, flags and library locations. 
! Be sure to link the program with both the P3DFFT library and the underlying 
! 1D FFT library such as ESSL or FFTW. 
!
! If you have questions please contact Dmitry Pekurovsky, dmitry@sdsc.edu

      program fft3d

      use p3dfft
      implicit none
      include 'mpif.h'

      integer i,n,nx,ny,nz
      integer m,x,y,z
      integer fstatus
      logical flg_inplace

      real(mytype), dimension(:,:,:,:),  allocatable :: BEG
      real(mytype) pi,twopi,sinyz,diff,cdiff,ccdiff,ans
      real(mytype), pointer :: pbeg(:,:,:)
      complex(mytype), pointer :: pend(:,:,:)


      integer*8 Ntot
      real(mytype) factor
      real(mytype),dimension(:),allocatable:: sinx,siny,sinz
      real(8) rtime1,rtime2
      real(8) t1,t2,t3,t4,gt1,gt2,gt3,gt4,tp1,gtp1
      common /timers/ t1,t2,t3,t4,tp1
      integer ierr,nu,ndim,dims(2),nproc,proc_id
      integer start(3),end(3),size(3)
      integer iproc,jproc

      real(8) starttime, finaltime

      call MPI_INIT (ierr)
      call MPI_COMM_SIZE (MPI_COMM_WORLD,nproc,ierr)
      call MPI_COMM_RANK (MPI_COMM_WORLD,proc_id,ierr)

      pi=atan(1.)*4.
      twopi=2.*pi

      t1 = 0.0
      t2 = 0.0
      t3 = 0.0
      t4 = 0.0
      tp1 = 0.0
      gt1=0.0
      gt2=0.0
      gt3=0.0
      gt4=0.0
      gtp1 = 0.0

      if (proc_id.eq.0) then 
         open (unit=3,file='stdin',status='old',
     &         access='sequential',form='formatted', iostat=fstatus)
         if (fstatus .eq. 0) then
            write(*, *) ' Reading from input file stdin'
         endif 
         ndim = 2

        read (3,*) nx, ny, nz, ndim,n, nu
        write (*,*) "procs=",nproc," nx=",nx,
     &          " ny=", ny," nz=", nz,"ndim=",ndim," repeat=", n,' nu=',nu
        if(mytype .eq. 4) then
           print *,'Single precision version'
        else if(mytype .eq. 8) then
           print *,'Double precision version'
        endif
       endif

      call MPI_Bcast(nx,1, MPI_INTEGER,0,mpi_comm_world,ierr)
      call MPI_Bcast(ny,1, MPI_INTEGER,0,mpi_comm_world,ierr)
      call MPI_Bcast(nz,1, MPI_INTEGER,0,mpi_comm_world,ierr)
      call MPI_Bcast(n,1, MPI_INTEGER,0,mpi_comm_world,ierr)
      call MPI_Bcast(nu,1, MPI_INTEGER,0,mpi_comm_world,ierr)
      call MPI_Bcast(ndim,1, MPI_INTEGER,0,mpi_comm_world,ierr)

!    nproc is divided into a iproc x jproc grid
!

      if(ndim .eq. 1) then
         dims(1) = 1
         dims(2) = nproc
      else if(ndim .eq. 2) then
         dims(1) = nproc / min(nx,ny)
         if(dims(1) .eq. 0) then
            dims(1) = 1
         endif
         dims(2) = nproc / dims(1)
      endif

      iproc = dims(1)
      jproc = dims(2)

      if(proc_id .eq. 0) then
         print *,'Using processor grid ',iproc,' x ',jproc
      endif

      call p3dfft_setup (dims,nx,ny,nz)
      call get_dims(start,end,size,1)

      allocate (sinx(nx))
      allocate (siny(ny))
      allocate (sinz(nz))
c
c initialize
c
      do z=start(3),end(3)
         sinz(z)=sin((z-1)*twopi/nz)
      enddo
      do y=start(2),end(2)
         siny(y)=sin((y-1)*twopi/ny)
      enddo
      do x=start(1),end(1)
         sinx(x)=sin((x-1)*twopi/nx)
      enddo

      allocate (BEG(start(1):end(1),start(2):end(2),start(3):end(3),nu), stat=ierr)
      if(ierr .ne. 0) then
         print *,'Error ',ierr,' allocating array BEG'
      endif

c start with x-z slabs in physical space
c
      do i=1,nu
         do z=start(3),end(3)
            do y=start(2),end(2)
               sinyz=siny(y)*sinz(z)
               do x=start(1),end(1)
                  BEG(x,y,z,i)=sinx(x)*sinyz 
               enddo
            enddo
         enddo
      enddo

c
c transform from physical space to wavenumber space
c (XgYiZj to XiYjZg)
c then transform back to physical space
c (XiYjZg to XgYiZj)
c

      factor = 1.0d0/(nx*ny)
      factor = factor / nz
      Ntot = size(1)*size(2)*size(3)/2
      flg_inplace = .true.

      starttime = MPI_Wtime()
      rtime1 = 0.0               
      do  m=1,n
         do i=1,nu
            if(proc_id .eq. 0) then
               print *,'Iteration ',m,', variable ',i
            endif

            rtime1 = rtime1 - MPI_wtime()
            call p3dfft_ftran_r2c (BEG(:,:,:,i),BEG(:,:,:,i),flg_inplace)

            rtime1 = rtime1 + MPI_wtime()

            call print_all(BEG(:,:,:,i),Ntot)

            call mult_array(BEG(:,:,:,i),
     &            Ntot,factor)

            rtime1 = rtime1 - MPI_wtime()
            call p3dfft_btran_c2r (BEG(:,:,:,i),BEG(:,:,:,i),
     &            flg_inplace)       
            rtime1 = rtime1 + MPI_wtime()

         end do
      enddo
      finaltime = MPI_Wtime()

      call MPI_Reduce(rtime1,rtime2,1,mpi_real8,MPI_MAX,0,
     &  MPI_COMM_WORLD,ierr)
     
      if (proc_id.eq.0) write(6,*)'proc_id, cpu time per loop',
     &   proc_id,rtime2/dble(n *nu)

      call MPI_Reduce(t1,gt1,1,mpi_real8,MPI_MAX,0,
     &  MPI_COMM_WORLD,ierr)
      call MPI_Reduce(tp1,gtp1,1,mpi_real8,MPI_MAX,0,
     &  MPI_COMM_WORLD,ierr)
      call MPI_Reduce(t2,gt2,1,mpi_real8,MPI_MAX,0,
     &  MPI_COMM_WORLD,ierr)
      call MPI_Reduce(t3,gt3,1,mpi_real8,MPI_MAX,0,
     &  MPI_COMM_WORLD,ierr)
      call MPI_Reduce(t4,gt4,1,mpi_real8,MPI_MAX,0,
     &  MPI_COMM_WORLD,ierr)

      gt1 = gt1 / dble(n*nu)
      gtp1 = gtp1 / dble(n*nu)
      gt2 = gt2 / dble(n*nu)
      gt3 = gt3 / dble(n*nu)
      gt4 = gt4 / dble(n*nu)
      
      if(proc_id .eq. 0) then
         print *,'t1,tp1,t2,t3,t4: ',gt1,gtp1,gt2,gt3,gt4
      endif

      cdiff=0.0d0
      do 20 z=start(3),end(3)
         do 20 y=start(2),end(2)
            sinyz=siny(y)*sinz(z)
            do 20 x=start(1),end(1)
            ans=sinx(x)*sinyz
            if(cdiff .lt. abs(BEG(x,y,z,1)-ans)) then
               cdiff = abs(BEG(x,y,z,1)-ans)
c               print *,'x,y,z,cdiff=',x,y,z,cdiff
            endif
 20   continue
      call MPI_Reduce(cdiff,ccdiff,1,mpireal,MPI_MAX,0,
     &  MPI_COMM_WORLD,ierr)
     
      if (proc_id.eq.0) then
        write (6,*) 'max diff at do 20=',ccdiff
        write (6,*) 'Time for fft =', finaltime - starttime
      endif


      call MPI_FINALIZE (ierr)

      end program

!=========================================================

      subroutine mult_array(X,nar,f)

      use p3dfft

      integer*8 nar,i
      complex(mytype) X(*)
      real(mytype) f

      do i=1,nar
         X(i) = X(i) * f
      enddo

      return
      end subroutine
      

      subroutine print_all(Ar,Nar)

      use p3dfft

      integer x,y,z
      integer(8) i,Nar
      complex(mytype) Ar(1,1,*)
      integer Fstart(3),Fend(3),Fsize(3)
      
      call get_dims(Fstart,Fend,Fsize,2)
      do i=1,Nar
         if(abs(Ar(1,1,i)) .gt. 0.5) then
            z = i/(Fsize(1)*Fsize(2))
            y = (i - z * Fsize(1)*Fsize(2))/Fsize(1)
            x = i-z*Fsize(1)*Fsize(2) - y*Fsize(1)
            print *,'(',x+1,y+1,z+1,') ',Ar(1,1,i)
         endif
      enddo

      return
      end subroutine

