#include <stdio.h>
#include "mpi.h"
#include "ia2a.h"

#define IA2A_TAG 9191

IA2A_Data_t  *IA2A_Data = NULL;

void IA2A_Init( MPI_Datatype sendtype, MPI_Datatype recvtype,
                MPI_Comm comm, int a2a_count )
{
    if ( IA2A_Data == NULL ) {
        IA2A_Data = (IA2A_Data_t *) malloc( sizeof(IA2A_Data_t) );
        if ( IA2A_Data == NULL ) {
            fprintf( stderr, "IA2A_Init() - malloc(IA2A_Data_t) fails.\n" );
            fflush( stderr );
            MPI_Abort( MPI_COMM_WORLD, -1 );
        }
        IA2A_Data->req_array  = NULL;
        IA2A_Data->stat_array = NULL;
    }

    IA2A_Data->comm        = comm;
    IA2A_Data->sendtype    = sendtype;
    IA2A_Data->recvtype    = recvtype;
    IA2A_Data->a2a_count   = a2a_count;

    MPI_Type_extent( IA2A_Data->sendtype, &(IA2A_Data->send_extent) );
    MPI_Type_extent( IA2A_Data->recvtype, &(IA2A_Data->recv_extent) );

    MPI_Comm_size( IA2A_Data->comm, &(IA2A_Data->comm_size) );
    MPI_Comm_rank( IA2A_Data->comm, &(IA2A_Data->comm_rank) );

    /* Factor of 2, 1 for Isend, 1 for Irecv */
    IA2A_Data->req_max   = 2 * IA2A_Data->comm_size * IA2A_Data->a2a_count;
    IA2A_Data->req_array = (MPI_Request *) malloc( IA2A_Data->req_max
                                                 * sizeof(MPI_Request) );
    if ( IA2A_Data->req_array == NULL ) {
        fprintf( stderr, "IA2A_Init() - malloc(req_array) fails.\n" );
        fflush( stderr );
        MPI_Abort( MPI_COMM_WORLD, -1 );
    }
    IA2A_Data->stat_array = (MPI_Status *) malloc( 2 * IA2A_Data->comm_size  
                                                 * sizeof(MPI_Status) );
    if ( IA2A_Data->stat_array == NULL ) {
        fprintf( stderr, "IA2A_Init() - malloc(stat_array) fails.\n" );
        fflush( stderr );
        MPI_Abort( MPI_COMM_WORLD, -1 );
    }

    IA2A_Data->req_count = 0;
}

void IA2A_Finalize( void )
{
    if ( IA2A_Data != NULL ) {
        if ( IA2A_Data->req_array != NULL ) {
            free( IA2A_Data->req_array );
            IA2A_Data->req_array = NULL;
        }
        if ( IA2A_Data->stat_array != NULL ) {
            free( IA2A_Data->stat_array );
            IA2A_Data->stat_array = NULL;
        }
        free( IA2A_Data );
        IA2A_Data = NULL;
    }
}
                

/*
   All arguments are similar to MPI_Alltoallv() except req_array
   which is an array of MPI_Request of size 2 * communicator's size.
*/
void IA2A_Post( void *sendbuf, int *sendcnts, int *sdispls, 
                void *recvbuf, int *recvcnts, int *rdispls )
{
    MPI_Comm     comm;
    MPI_Datatype sendtype, recvtype;
    MPI_Aint     send_extent, recv_extent;
    int          comm_rank, comm_size;
    int          dest, idx, ierr;
  
    comm        = IA2A_Data->comm;
    sendtype    = IA2A_Data->sendtype;
    recvtype    = IA2A_Data->recvtype;
    send_extent = IA2A_Data->send_extent;
    recv_extent = IA2A_Data->recv_extent;
    comm_rank   = IA2A_Data->comm_rank;
    comm_size   = IA2A_Data->comm_size;

    for ( idx = 0; idx < comm_size; idx++ ) {
        if ( IA2A_Data->req_count >= IA2A_Data->req_max ) {
            fprintf( stderr, "IA2A_Post: Max out request[] during Irecv!\n" );
            fflush( stderr );
            MPI_Abort( MPI_COMM_WORLD, -1 );
        }
        dest = (comm_rank + idx) % comm_size;
        ierr = MPI_Irecv( (char *) recvbuf + rdispls[dest] * recv_extent,
                          recvcnts[dest], recvtype, dest, IA2A_TAG, comm,
                          &(IA2A_Data->req_array[IA2A_Data->req_count]) );
        if ( ierr != MPI_SUCCESS ) {
            fprintf( stderr, "IA2A_Post: Irecv error!\n" );
            fflush( stderr );
            MPI_Abort( MPI_COMM_WORLD, -1 );
        }
        IA2A_Data->req_count++;
    }

    for ( idx = 0; idx < comm_size; idx++ ) {
        if ( IA2A_Data->req_count >= IA2A_Data->req_max ) {
            fprintf( stderr, "IA2A_Post: Max out request[] during Isend!\n" );
            fflush( stderr );
            MPI_Abort( MPI_COMM_WORLD, -1 );
        }
        dest = (comm_rank + idx) % comm_size;
        ierr = MPI_Isend( (char *) sendbuf + sdispls[dest] * send_extent,
                          sendcnts[dest], sendtype, dest, IA2A_TAG, comm,
                          &(IA2A_Data->req_array[IA2A_Data->req_count]) );
        if ( ierr != MPI_SUCCESS ) {
            fprintf( stderr, "IA2A_Post: Isend errorr!\n" );
            fflush( stderr );
            MPI_Abort( MPI_COMM_WORLD, -1 );
        }
        IA2A_Data->req_count++;
    }
}

void IA2A_Fence( void )
{
    int ierr, idx, jdx, isr_count;
    isr_count = 2 * IA2A_Data->comm_size;
    for ( idx = 0; idx < IA2A_Data->a2a_count; idx++ ) {
        ierr = MPI_Waitall( isr_count,
                            &(IA2A_Data->req_array[idx*isr_count]),
                            IA2A_Data->stat_array );
        if ( ierr == MPI_ERR_IN_STATUS ) {
            for ( jdx = 0; jdx < isr_count; jdx++ ) {
                if ( IA2A_Data->stat_array[jdx].MPI_ERROR != MPI_SUCCESS ) {
                    fprintf( stderr, "IA2A_Fence: Wait error!\n");
                    fflush( stderr );
                    MPI_Abort( MPI_COMM_WORLD, -1 );
                }
            }
        }
    }
}
