
! This sample program illustrates the 
! use of P3DFFT library for highly scalable parallel 3D FFT. 
!
! This program initializes a 3D array with a 3D sine wave, then 
! performs forward transform, backward transform, and checks that 
! the results are correct, namely the same as in the start except 
! for a normalization factor. It can be used both as a correctness
! test and for timing the library functions. 
!
! To build this program, use one of the provided makefiles, modifying 
! it as needed for your compilers, flags and library locations. 
! Be sure to link the program with both the P3DFFT library and the underlying 
! 1D FFT library such as ESSL or FFTW. 
!
! If you have questions please contact Dmitry Pekurovsky, dmitry@sdsc.edu

      program fft3d

      use p3dfft
      implicit none
      include 'mpif.h'
c      include 'p3dfft_incl_single.f'


      integer i,n,nx,ny,nz
      integer m,x,y,z
      integer fstatus
      logical flg_inplace

      real(mytype), dimension(:,:,:,:),  allocatable :: BEG
      complex(mytype), dimension(:,:,:,:),  allocatable :: END
      real(mytype) pi,twopi,sinyz,diff,cdiff,ccdiff,ans

      integer*8 Ntot
      real(mytype) factor
      real(mytype),dimension(:),allocatable:: sinx,siny,sinz
      real(8) rtime1,rtime2
      real(8) t1,t2,t3,t4,gt1,gt2,gt3,gt4,tp1,gtp1
      common /timers/ t1,t2,t3,t4,tp1
      integer ierr,nu,ndim,dims(2),nproc,proc_id
      integer xst,xen,xsz,yst,yen,ysz,zst,zen,zsz
      integer Fxst,Fxen,Fxsz,Fyst,Fyen,Fysz,Fzst,Fzen,Fzsz
      integer iproc,jproc

      call MPI_INIT (ierr)
      call MPI_COMM_SIZE (MPI_COMM_WORLD,nproc,ierr)
      call MPI_COMM_RANK (MPI_COMM_WORLD,proc_id,ierr)

      pi=atan(1.)*4.
      twopi=2.*pi

      t1 = 0.0
      t2 = 0.0
      t3 = 0.0
      t4 = 0.0
      tp1 = 0.0
      gt1=0.0
      gt2=0.0
      gt3=0.0
      gt4=0.0
      gtp1 = 0.0

      if (proc_id.eq.0) then 
         open (unit=3,file='stdin',status='old',
     &         access='sequential',form='formatted', iostat=fstatus)
         if (fstatus .eq. 0) then
            write(*, *) ' Reading from input file stdin'
         endif 
         ndim = 2

        read (3,*) nx, ny, nz, ndim,n, nu
        write (*,*) "procs=",nproc," nx=",nx,
     &          " ny=", ny," nz=", nz,"ndim=",ndim," repeat=", n,' nu=',nu
        if(mytype .eq. 4) then
           print *,'Single precision version'
        else if(mytype .eq. 8) then
           print *,'Double precision version'
        endif
       endif

      call MPI_Bcast(nx,1, MPI_INTEGER,0,mpi_comm_world,ierr)
      call MPI_Bcast(ny,1, MPI_INTEGER,0,mpi_comm_world,ierr)
      call MPI_Bcast(nz,1, MPI_INTEGER,0,mpi_comm_world,ierr)
      call MPI_Bcast(n,1, MPI_INTEGER,0,mpi_comm_world,ierr)
      call MPI_Bcast(nu,1, MPI_INTEGER,0,mpi_comm_world,ierr)
      call MPI_Bcast(ndim,1, MPI_INTEGER,0,mpi_comm_world,ierr)

!    nproc is devided into a iproc x jproc stencle
!

      if(ndim .eq. 1) then
         dims(1) = 1
         dims(2) = nproc
      else if(ndim .eq. 2) then
         dims(1) = 0
         dims(2) = 0
         call MPI_Dims_create(nproc,2,dims,ierr)
         if(dims(1) .gt. dims(2)) then
            dims(1) = dims(2)
            dims(2) = nproc / dims(1)
         endif
      endif

      iproc = dims(1)
      jproc = dims(2)

      if(proc_id .eq. 0) then
         print *,'Using processor grid ',iproc,' x ',jproc
      endif

      call p3dfft_setup (dims,nx,ny,nz)
      call get_dims(xst,xen,xsz,yst,yen,ysz,zst,zen,zsz,1)
      call get_dims(Fxst,Fxen,Fxsz,Fyst,Fyen,Fysz,Fzst,Fzen,Fzsz,2)

      allocate (sinx(nx))
      allocate (siny(ny))
      allocate (sinz(nz))
c
c initialize
c
      do z=zst,zen
         sinz(z)=sin((z-1)*twopi/nz)
      enddo
      do y=yst,yen
         siny(y)=sin((y-1)*twopi/ny)
      enddo
      do x=xst,xen
         sinx(x)=sin((x-1)*twopi/nx)
      enddo

      allocate (BEG(xst:xen,yst:yen,zst:zen,nu), stat=ierr)
      if(ierr .ne. 0) then
         print *,'Error ',ierr,' allocating array BEG'
      endif
      allocate (END(Fxst:Fxen,Fyst:Fyen,Fzst:Fzen,nu), stat=ierr)
      if(ierr .ne. 0) then
         print *,'Error ',ierr,' allocating array END'
      endif

c start with x-z slabs in physical space
c
      do i=1,nu
         do z=zst,zen
            do y=yst,yen
               sinyz=siny(y)*sinz(z)
               do x=xst,xen
                  BEG(x,y,z,i)=sinx(x)*sinyz 
               enddo
            enddo
         enddo
      enddo

c
c transform from physical space to wavenumber space
c (XgYiZj to XiYjZg)
c then transform back to physical space
c (XiYjZg to XgYiZj)
c

      factor = 1.0d0/(nx*ny)
      factor = factor / nz
      Ntot = xsz*ysz*zsz/2
      flg_inplace = .true.

      rtime1 = 0.0               
      do  m=1,n
         do i=1,nu
            if(proc_id .eq. 0) then
               print *,'Iteration ',m,', variable ',i
            endif

            rtime1 = rtime1 - MPI_wtime()
            call p3dfft_ftran_r2c (BEG(:,:,:,i),END(:,:,:,i),flg_inplace)

            rtime1 = rtime1 + MPI_wtime()

            call print_all(END(Fxst,Fyst,Fzst,i))

            call mult_array(END(Fxst,Fyst,Fzst,i),
     &            Ntot,factor)

            rtime1 = rtime1 - MPI_wtime()
            call p3dfft_btran_c2r (END(Fxst,Fyst,Fzst,i),BEG(xst,yst,zst,i),
     &            flg_inplace)       
            rtime1 = rtime1 + MPI_wtime()

         end do
      enddo

c      rtime1 = MPI_wtime() - rtime1
      call MPI_Reduce(rtime1,rtime2,1,mpi_real8,MPI_MAX,0,
     &  MPI_COMM_WORLD,ierr)
     
      if (proc_id.eq.0) write(6,*)'proc_id, cpu time per loop',
     &   proc_id,rtime2/dble(n *nu)

      call MPI_Reduce(t1,gt1,1,mpi_real8,MPI_MAX,0,
     &  MPI_COMM_WORLD,ierr)
      call MPI_Reduce(tp1,gtp1,1,mpi_real8,MPI_MAX,0,
     &  MPI_COMM_WORLD,ierr)
      call MPI_Reduce(t2,gt2,1,mpi_real8,MPI_MAX,0,
     &  MPI_COMM_WORLD,ierr)
      call MPI_Reduce(t3,gt3,1,mpi_real8,MPI_MAX,0,
     &  MPI_COMM_WORLD,ierr)
      call MPI_Reduce(t4,gt4,1,mpi_real8,MPI_MAX,0,
     &  MPI_COMM_WORLD,ierr)

      gt1 = gt1 / dble(n*nu)
      gtp1 = gtp1 / dble(n*nu)
      gt2 = gt2 / dble(n*nu)
      gt3 = gt3 / dble(n*nu)
      gt4 = gt4 / dble(n*nu)
      
      if(proc_id .eq. 0) then
         print *,'t1,tp1,t2,t3,t4: ',gt1,gtp1,gt2,gt3,gt4
      endif

      cdiff=0.0d0
      do 20 z=zst,zen
         do 20 y=yst,yen
            sinyz=siny(y)*sinz(z)
            do 20 x=xst,xen
            ans=sinx(x)*sinyz
            if(cdiff .lt. abs(BEG(x,y,z,1)-ans)) then
               cdiff = abs(BEG(x,y,z,1)-ans)
c               print *,'x,y,z,cdiff=',x,y,z,cdiff
            endif
 20   continue
      call MPI_Reduce(cdiff,ccdiff,1,mpireal,MPI_MAX,0,
     &  MPI_COMM_WORLD,ierr)
     
      if (proc_id.eq.0) write (6,*) 'max diff at do 20=',ccdiff

      call MPI_FINALIZE (ierr)

      contains 
!=========================================================

      subroutine mult_array(X,nar,f)

      integer*8 nar,i
      complex(mytype) X(nar)
      real(mytype) f

      do i=1,nar
         X(i) = X(i) * f
      enddo

      return
      end subroutine
      

      subroutine print_all(Ar)

      integer x,y,z
      complex(mytype) Ar(Fxst:Fxen,Fyst:Fyen,Fzst:Fzen)

      do z=Fzst,Fzen
         do y=Fyst,Fyen
            do x=Fxst,Fxen
               if(abs(Ar(x,y,z)) .gt. 0.5) then
                  print *,'(',x,y,z,') ',Ar(x,y,z)
               endif
            enddo
         enddo
      enddo

      return
      end subroutine

      end program
